/* wwwfetch.c */

#include "config.h"

#if defined(HTEX) || defined(XHDVI)
#ifdef HAVE_NETDB_H
#include <netdb.h> /* for struct hostent */
#endif
#include <kpathsea/c-stat.h> /* libwww headers use struct stat */
#ifdef HTEX
#include "WWWLib.h"
#include "HTReqMan.h"
#include "HTUtils.h"
#include "HTInit.h"
#else
#include <stdio.h>
#include "patchlevel.h"
#include "url.h"
#endif
#include "HTAccess.h"
#include "HText.h"
#include "HTParse.h"
#include "HTFWrite.h"
#include "HTFormat.h"
#include "HTTP.h"
#include "HTAlert.h"

/* Given absolute URL, open a temporary filename, and do the transfer */

#define LINE 512
char buf[LINE];

PUBLIC char *HTAppName = "XHdvi"; /* Application name */
#ifdef HTEX
PUBLIC char *HTAppVersion = ""; /* Application version */
#else
PUBLIC char *HTAppVersion = PATCHLEVEL; /* Application version */
#endif

/* Following need to be defined for the WWW library, but are not used */

HText *  HText_new       (HTRequest * request, HTParentAnchor * anchor)
{
    fprintf(stderr, "HText_new called\n");
    return 0;
}

HText *  HText_new2      (HTRequest *            request,
                                 HTParentAnchor *       anchor,
                                 HTStream *             output_stream)
{
    fprintf(stderr, "HText_new2 called\n");
    return 0;
}

HText *  HText_new3      (HTRequest *            request,
                                 HTStream *             output_stream,
                                 HTStructured *         structure)
{
    fprintf(stderr, "HText_new3 called\n");
    return 0;
}

void     HText_free (HText * me)
{
    fprintf(stderr, "HText_free called\n");
   
}

void HText_beginAppend (HText * text)
{
    fprintf(stderr, "HText_beginAppend called\n");
   
}

void HText_endAppend (HText * text)
{
    fprintf(stderr, "HText_endAppend called\n");
    
}

void HText_setStyle (HText * text, HTStyle * style)
{
    fprintf(stderr, "HText_setStyle called\n");
    
}

void HText_appendCharacter (HText * text, char ch)
{
    fprintf(stderr, "HText_appendCharacter called\n");
    
}

void HText_appendText (HText * text, const char * str)
{
    fprintf(stderr, "HText_appendText called\n");
    
}

void HText_appendParagraph (HText * text)
{
    fprintf(stderr, "HText_appendParagraph called\n");
    
}

void HText_appendLineBreak (HText * text)
{
    fprintf(stderr, "HText_appendLineBreak called\n");
    
}

void HText_appendHorizontalRule (HText * text)
{
    fprintf(stderr, "HText_appendHorizontalRule called\n");
   
}

void HText_beginAnchor (HText * text, HTChildAnchor * anc)
{
    fprintf(stderr, "HText_beginAnchor called\n");
    
}

void HText_endAnchor (HText * text)
{
    fprintf(stderr, "HText_endAnchor called\n");
   
}

void HText_setNextId (
        HText *         text,
        const char *    s)
{
    fprintf(stderr, "HText_setNextId called\n");
   
}

void HText_appendImage (
        HText *         text,
        HTChildAnchor * anc,
        const char *    alternative_text,
        const char *    alignment,
        BOOL            isMap)
{
    fprintf(stderr, "HText_appendImage called\n");
    
}

void HText_appendObject (
        HText *                 text,
        int                     element_number,
        const BOOL *            present,
        const char * const *    value)
{
    fprintf(stderr, "HText_appendObject called\n");
    
}

void HText_unimplemented (
        HText *         text)
{
    fprintf(stderr, "HText_unimplemented called\n");
    
}

void HText_dump (HText * me)
{
    fprintf(stderr, "HText_dump called\n");
    
}

HTParentAnchor * HText_nodeAnchor (HText * me)
{
    fprintf(stderr, "HText_nodeAnchor called\n");
    return 0;
}

BOOL HText_select (HText * text)
{
    fprintf(stderr, "HText_select called\n");
    return 0;
}

BOOL HText_selectAnchor (HText * text, HTChildAnchor* anchor)
{
    fprintf(stderr, "HText_selectAnchor called\n");
    return 0;
}

void HText_applyStyle (HText * me, HTStyle *style)
{
    fprintf(stderr, "HText_applyStyle called\n");
    
}

void HText_updateStyle (HText * me, HTStyle *style)
{
    fprintf(stderr, "HText_updateStyle called\n");
    
}

HTStyle * HText_selectionStyle (
        HText * me,
        HTStyleSheet* sheet)
{
    fprintf(stderr, "HText_selectionStyle called\n");
    return 0;
}

void HText_replaceSel (HText * me,
        const char *aString,
        HTStyle* aStyle)
{
    fprintf(stderr, "HText_replaceSel called\n");
    
}

void HTextApplyToSimilar (HText * me, HTStyle *style)
{
    fprintf(stderr, "HTextApplyToSimilar called\n");
    
}

void HTextSelectUnstyled (HText * me, HTStyleSheet *sheet)
{
    fprintf(stderr, "HTextSelectUnstyled called\n");
    
}

void             HText_unlinkSelection (HText * me)
{
    fprintf(stderr, "HText_unlinkSelection called\n");
    
}

HTAnchor *       HText_referenceSelected (HText * me)
{
    fprintf(stderr, "HText_referenceSelected called\n");
    return 0;
}

HTAnchor *       HText_linkSelTo (HText * me, HTAnchor* anchor)
{
    fprintf(stderr, "HText_linkSelTo called\n");
    return 0;
}




HTStyleSheet *styleSheet = NULL;
HText *HTMainText = NULL;
HTParentAnchor *HTMainAnchor = NULL;

/* Given absolute URL, open a temporary filename, and do the transfer */

int
www_fetch(url, savefile)
char *url;
char *savefile;
{
	int resp;
	HTRequest *reqptr;
	FILE *fp;

	if ((fp = fopen(savefile, "w")) == NULL) {
		fprintf(stderr, "Cannot open %s for writing\n", savefile);
		exit(0); /* No cleanup! */
	}
	reqptr = HTRequest_new(); /* A blank request with defaults set */
#if 0
/* Doesn't exist in the libwww I got as of 7sep95 --karl. */
	HTEnableFrom = YES; /* Says who we are - e-mail address */
#endif
	reqptr->output_format = WWW_SOURCE; /* Write out as it came? */
/* Might want HTSaveAndCallBack? */
	reqptr->output_stream = HTFWriter_new(reqptr, fp, NO);
	HTFormatInit(reqptr->conversions); /* Not sure what this does... */
	resp = HTLoadAbsolute(url, reqptr);
/* We're going to have to figure out what's in the file somehow... */
/*	printf("Content-type: %s\n", (reqptr->content_encoding)->name); */
#ifdef DOFORKS
	exit(1); /* No cleanup! */
#else
	return resp;
#endif
}

/* Turn a relative URL into an absolute one: */
void
make_absolute(rel, base, len)
    char *rel, *base;
    int len;
{
        char *cp, *parsed;

	if (base == NULL) return;
        cp = strchr(rel, '\n');
        if (cp) *cp = '\0'; /* Eliminate newline char */

        parsed = HTParse(rel, base, PARSE_ALL);
        strncpy(rel, parsed, len);
        free(parsed);
}

/* buf is a URL - savefile is local file name to save the file in */
static void
make_tmp_filename(tmpdir, buf, savefile)
    char *tmpdir, *buf, **savefile;
{
	char *cp, *cp2;
	int i, j, fd;

	*savefile = NULL;
	cp = strrchr(buf, '/');
	if (cp == NULL) cp = strrchr(buf, ':');
	if (cp == NULL) return; /* Unsuccessful... */
	if (strlen(cp) < 3) cp = "/tmpfile";
	cp++;
	*savefile = (char *) malloc((strlen(cp) + strlen(tmpdir)+3)*sizeof(char));
	cp2 = strrchr(cp, '.'); /* file extension */
	if (cp2 != NULL) {
		*cp2 = '\0';
		cp2++;
		for (i=0; i < 26; i++) { /* Attempt to create new file */
			sprintf(*savefile,"%s/%s%c.%s", tmpdir,cp,'a'+i,cp2);
			for (j=0; j < nURLs; j++) {
				if (!strcmp(*savefile, filelist[j].file)) break;
			}
			if (j < nURLs) continue;
/*
			if ((fd = creat(*savefile,0600)) >= 0) {
*/
			if ((fd = creat(*savefile,0666)) >= 0) {
				close(fd);
				break;
			}
		}
		cp2--;
		*cp2 = '.'; /* Undo the damage we just did to buf... */
	} else {
		for (i=0; i < 26; i++) { /* Attempt to create new file */
			sprintf(*savefile,"%s/%s%c", tmpdir,cp,'a'+i);
			for (j=0; j < nURLs; j++) {
				if (!strcmp(*savefile, filelist[j].file)) break;
			}
			if (j < nURLs) continue;
			if ((fd = creat(*savefile,0600)) >= 0) {
				close(fd);
				break;
			}
		}
	}
	if (i == 26) { /* Unsuccessful */
		free(*savefile);
		*savefile = NULL;
	}
}


typedef struct {
	char *url;
	char *savefile;
	pid_t childnum;
} Fetch_Children;

#define MAXC 10 /* Don't send off more than 10 simultaneously */
Fetch_Children fetch_children[MAXC];
static int nchildren = MAXC+1;

void
wait_for_urls() /* Wait for all the children to finish... */
{
	int i;

	if (nchildren > MAXC) {
		/* Initialization needed: */
		for (i=0; i < MAXC; i++) {
			fetch_children[i].url = NULL;
			fetch_children[i].savefile = NULL;
		}
	} else {
		for (i=0; i < nchildren; i++) {
#ifdef DOFORKS
			ret = wait(&status); /* Wait for one to finish */
			for (j=0; j < nchildren; j++) {
				if (ret == fetch_children[j].childnum) {
					/* testing: */
					printf("URL %s in file %s: status %d\n",
						fetch_children[j].url,
						fetch_children[j].savefile,
						status);
					break;
				}
			}
#else
			printf("URL %s in file %s\n",
				fetch_children[i].url,
				fetch_children[i].savefile);
#endif
		}
	}
	nchildren = 0;
}

/* Start a fetch of a relative URL */
int
fetch_relative_url(base_url, rel_url, tmpdir)
    char *base_url;
    const char *rel_url;
    char *tmpdir;
{
	int i, resp;
	char *savefile;
	char *cp;

	/* Step 1: make the relative URL absolute: */
        strncpy(buf, rel_url, LINE); /* Put it in buf */
	make_absolute(buf, base_url, LINE);

	/* Step 1.5: Check whether we already have it - if so return */
	for (i=0; i < nURLs; i++) {
		if (!strcmp(buf, filelist[i].url)) return(i);
	}
	
	/* Step 2: Find a temporary file to store the output in */
	make_tmp_filename(tmpdir, buf, &savefile);
	if (savefile == NULL) {
		fprintf(stderr, "Could not find temporary file for %s in %s\n",
			buf, tmpdir);
		return(-1);
	}

	/* Step 3: Fork to fetch the URL */
	if (nchildren >= MAXC) wait_for_urls(); /* Wait for the old ones */
	cp = NULL;
	fetch_children[nchildren].url = MyStrAllocCopy(&cp, buf);
	cp = NULL;
	fetch_children[nchildren].savefile = MyStrAllocCopy(&cp, savefile);

	/* Step 4: Update the URL-filename list */
	if (nURLs == 0) {
		filelist = (FiletoURLconv *) xmalloc(FILELISTCHUNK*sizeof(FiletoURLconv), "filelist");
		bzero(filelist,  FILELISTCHUNK*sizeof(FiletoURLconv));
	} else if (nURLs%FILELISTCHUNK == 0) {
		filelist = (FiletoURLconv *) realloc(filelist, (nURLs+FILELISTCHUNK)*sizeof(FiletoURLconv));
		bzero(filelist + nURLs, FILELISTCHUNK*sizeof(FiletoURLconv));
	}
	MyStrAllocCopy(&(filelist[nURLs].url), buf);
	MyStrAllocCopy(&(filelist[nURLs].file), savefile);
	nURLs++;
#ifdef DOFORKS
	fetch_children[nchildren].childnum = fork();
	if (fetch_children[nchildren].childnum == 0) {  /* Child process */
		www_fetch(buf, savefile); /* Get the URL! */
		exit(0); /* Make sure this process quits... */
	}
	nchildren++;
#else
	resp = www_fetch(buf, savefile); /* Get the URL! */
	if (resp == 0) {
		paint_anchor("Error: Can't locate URL!\n");
		nURLs--;
		unlink(filelist[nURLs].file); /* Get rid of that temp file */
		return(-1);
	} else {
		nchildren++;
	}
#endif
	return(nURLs-1);
}


#endif /* HTEX || XHDVI */

